# -*- coding: utf-8 -*-
"""
Modul zur Abfragen und Bearbeitung von Datenbanktabelle.

Autor:      W. Nagy
Startdatum: 16.12.2010
Lizenz:     http://creativecommons.org/licenses/by/2.0/at/

Die Daten eines Datensatzes werden in einer Domain gespeichert.

Die Domain kann dazu verwendet werden, Datensatzspezifische
Funktionen als Methode zu halten. z.B. ORM
Naehere Beschreibung am ende der Datei.

Dieses Modul ist fuer sqlite Datenbanken optimiert.

Es bietet die Grundlegenden Funktionen
- Oeffnen der Datenbank
- lesen eines Datensatzes ueber seinen Primary Key
  oder ein beliebige where Klausel
- Itteration von Datensaetzen (eachDomain)
  Mit der Moeglichkeit einen Filter einzusetzen und das Sortierkriterium anzugeben
- Einfuegen eines Datensatzes (insert)
- Aenderung eines Datensatzes (update)
- SQL Aggregatfunktionen (min, max, avg, sum, count)



ToDo:
   mySQL Anbindung

Update:


"""

import sqlite3
from dateconverter import Dateconverter
from checktype import Checktype

__version__='0.5'

class DbSqlite3 :
   """SQL Datenbankverbindung spezialisiert fuer SQLite

      HINT:
         Die Datenbank wird automatisch auf Autocommit gesetzt

      """
   
   connection = None
   
   def __init__(self,filename) :
      '''Konsturktor: fuer sqlite ist nur der Datenbankpfad notwendig.
         @param   filename       Dateiname und Pfad auf die Datenbankdatei.
         '''
      
      self.connection = sqlite3.connect(filename)
      self.connection.text_factory = str
      self.connection.isolation_level = None

   def __del__(self):
      self.connection.close()

class Database :
   """Allgemeines Datenbank Objekt."""
   
   db = None
   filename = ''

   def __init__(self,dbtype,*args) :
      """Initialisierung der Datenbank.
         @param dbtype:   Fuer jede vorhanden Datenbank wird ein Type angegeben.
         @param args:     Parameterliste:
                          fuer unterschiedliche Datenbanken werden unterschiedliche
                          Initialisierungstypen verwendet.
         """

      if dbtype == 'sqlite' :
         self.filename = args[0]
         self.db = DbSqlite3(args[0])
      else :
         raise "Ungueltiger Datenbanktype '%s'" % (dbtype)
         
   
class EachDomain() :
   """ Durchlaufen eines Datenbankstroms.
       Optional koennen bei der Instanzierung eine Kriterien und eine
       Sortierung kodieren.
       
       Die Daten werden nach dem Lesen in die Domaifelder
       kopiert.
       
       Usage:
         for lovDom in lov.eachDomain(where='lovID=23','lovClass')
            print lovDom
            
       """
       
   db = None
   cursor = None
   domain = None

   def __init__(self,domain,where,orderby) :
      """Iteration ueber Domain initialisieren

         @param domain    Domainobjekt
         @param where     (Optional) where Klausel fuer SQL Anweisung
         @param orderby   (Optional) Sortierklausel
      
         """

      self.domain = domain
      
      self.cursor = domain.db.db.connection.cursor()
      
      whereText = ''
      orderbyText = ''
      
      if where != None :
         whereText = ' where '+where
         
      if orderby != None :
         orderbyText = ' order by '+orderby
         
      sql = 'select * from '+domain.meta['tablename']+whereText+orderbyText
      
      self.cursor.execute(sql)
      self.value = None
      
   def __iter__(self) :
      'Gibt Iterator zurueck.'
      return self
   
   def next(self) :
      """durchlaufen des Datenstromes bis EOR auftritt. """
      record = self.cursor.fetchone()

      if record == None :
         raise StopIteration
      
      fieldnames = self.domain.getFieldnames(self.cursor)

      for fld in self.domain.meta['fields'] :
         iFieldnames  = fieldnames.index(fld)
         self.domain.__dict__[fld] = record[iFieldnames]

      return self.domain


class GetDomain() :
   """Liefert genau einen Datensatz aus 
      einer Tabelle und gibt die Daten in die Domainfelder
      """
   domain = None
   
   def __init__(self,domain) :
      """Initalisieren:
          @param domain Domain
          """

      self.domain = domain
      
      
   def get(self,domain,where) :
      primkey = domain.meta['primarykey']
      return primkey

      


class Domain(Checktype) :
   
   db                   = None
   cursor               = None
   eachDomain           = None
   tablename            = None
   currec               = {}
   isOk                 = False
   typecheckNoneAllowed = True          # True bei Typencheck ist None ein erlaubter Wert
   lastsql              = None
   typecheckStrict      = True          # False schaltet Typenueberpruefung ab

   def raiseTypeError(self,fldtype,name,value):
      """gibt Fehlermeldung bei Typenfehler aus"""
      myType =  str(type(value))
      raise TypeError('"%(value)s" ist nicht vom Typ %(type)s in Feld %(name)s, Typ: %(mytype)s in Klasse: "(%(classname)s)"' % {'value':str(value),'type':fldtype,'name':name,'classname':self.__class__,'mytype':myType})

   def __setattr__(self,name,value):
      """Pruefen auf Typengueltigkeit """

      # Wurde Typenueberpruefung abgeschalten
      if not self.typecheckStrict:
         self.__dict__[name] = value
         return

      # None als Eingabewert erlaubt.
      if self.typecheckNoneAllowed and value == None:
         self.__dict__[name] = value
         return
         
      # Ist das ein Domainfeld
      if name in self.meta['fields']:
         # Feldobjekt
         fld = self.meta['fields'][name]
         # Wurde type in meta deklariert
         if 'type' in fld:
            # Integer pruefen
            if fld['type'] == 'Integer':
               if self.isInteger(value): 
                  self.__dict__[name] = value
               else: 
                  self.raiseTypeError(fld['type'],name,value)

            # String pruefen
            elif fld['type'] == 'String':
               if self.isString(value): 
                  self.__dict__[name] = value
               else: 
                  self.raiseTypeError(fld['type'],name,value)

            # Emailadresse pruefen
            elif fld['type'] == 'Email':
               if self.isString(value): 
                  self.__dict__[name] = value
               else: 
                  self.raiseTypeError(fld['type'],name,value)
            # Float
            elif fld['type'] == 'Float':
               if self.isFloat(value): 
                  self.__dict__[name] = value
               else: 
                  self.raiseTypeError(fld['type'],name,value)
            # Date
            elif fld['type'] == 'Date':
               if self.isDate(value): 
                  dc = Dateconverter()
                  self.__dict__[name] = dc.giveAsANSIDateTime(value,secs=True)
               else: 
                  self.raiseTypeError(fld['type'],name,value)
            else:
               pass
         else:
            pass
      else:
         self.__dict__[name] = value

   def __init__(self,db) :
      'Initialisierung der Domainklasse.'
      self.db = db
      if self.db != None: self.isOk = True
      self.clear()
      self.cursor = db.db.connection.cursor()

   def getFieldnames(self,cursor) :
      'Liefert eine Liste mit Feldname fuer den Cursor'
      lstField = []
      
      for item in cursor.description :
         lstField.append(item[0])
      
      return lstField

   def set(self,name=None,value=None):
      """ Setzt den Inhalt eines Feldes der Domain"""
      if name == None:
         raise ValueError("Domain:: Feldname fehlt")
      self.__setattr__(name,value)

   def eachDomain(self,where=None,orderby=None) :
      """Iterator fuer jedes Vorkommen in der Datenbanktabelle.
      
         @param where       (Optional) Where Klauses fuer die SQL Anwesung
         @param orderby     (Optional) Sortierklausel
         
         usage
            for dom in domain.eachDomain() :
               print dom
         """
      
      return EachDomain(self,where,orderby)

   def clear(self):
      """setzt alle Attribute auf None"""
      for fld in self.meta['fields'].keys():
         self.__dict__[fld] = None
      

   def get(self,id=None,where=None):
      """Setzt die Domain auf den angegebnen Datensatz"""

      if id==None and where==None:
         raise Exception("get: Keine Parameter angegeben")

      primkey     = self.meta['primarykey']
      tablename   = self.meta['tablename']
      self.cursor = self.db.db.connection.cursor()

      if (id != None):
         select = 'select * from %(tablename)s where %(pk)s = ?' % {'tablename':tablename,'pk':primkey}
         self.cursor.execute(select,[id])
      else:
         select = 'select * from %(tablename)s where %(where)s'% {'tablename':tablename,'where':where}
         self.cursor.execute(select)
         

      fieldnames = self.getFieldnames(self.cursor)

      record = self.cursor.fetchone()
      if record != None:
         self.isOk = True
         for fld in self.meta['fields'] :
            iFieldnames  = fieldnames.index(fld)
            self.__dict__[fld] = record[iFieldnames]
      else:
         self.isOk = False
         for fld in self.meta['fields'] :
            iFieldnames  = fieldnames.index(fld)
            self.__dict__[fld] = None

   def delete(self):
      """loeschen aktueller Datensatz"""
      primkey     = self.meta['primarykey']
      tablename   = self.meta['tablename']
      self.cursor = self.db.db.connection.cursor()
      id = self.__dict__[primkey]
      if id==None:
         raise Exception("Aktueller Datensatz hat keine Eintrag in %(primkey)s" % {'primkey':primkey})
      
      self.cursor.execute('delete from %(tablename)s where %(primkey)s = %(id)s' % {'tablename':tablename,'primkey':primkey,'id':id})

   def deleteAll(self,where=None):
      """Loeschen basierend auf einet Where Klausel"""

      if where == None:
         raise Exception("bei delteAll ist Parameter where unbedingt notwendig")

      tablename   = self.meta['tablename']
      self.cursor = self.db.db.connection.cursor()
      self.cursor.execute("delete from %(tablename)s where %(where)s" % {'tablename':tablename,'where':where})
      
   def update(self):
      """Veraendern des Datensatzes basieren auf den Inhalten der Domain"""
      flds = list()
      values = list()
      primkey     = self.meta['primarykey']

      for fld in self.meta['fields'].keys():
         flds.append(fld)
         if self.__dict__[fld] == None:
            values.append(None)
         else:
            values.append(self.__dict__[fld])
      tablename   = self.meta['tablename']
      fragezeichen = '?,' * len(flds)
      fragezeichen = fragezeichen[:-1]

      fldList = list()

      for fld in flds:
         fldList.append('%(fld)s=?' % {'fld':fld})
      values.append(self.__dict__[primkey])

      stm = 'update %(tablename)s set %(flds)s where %(primkey)s = ?' % {
         'tablename':tablename,
         'flds':','.join(fldList),
         'primkey':primkey
         }
      self.lastsql = stm
      self.cursor.execute(stm,values)
      

   def insert(self):
      """Einfuegen eines Datensatzes
         Die aktuellen Werte aus der Domain werden in 
         die Datenbank geschrieben.
         """
      flds = list()
      values = list()

      for fld in self.meta['fields'].keys():
         flds.append(fld)
         if self.__dict__[fld] == None:
            values.append(None)
         else:
            values.append(self.__dict__[fld])
      tablename   = self.meta['tablename']
      fragezeichen = '?,' * len(flds)
      fragezeichen = fragezeichen[:-1]

      stm = 'insert into %(tablename)s (%(flds)s) values(%(fragezeichen)s)' % {
         'tablename':tablename,
         'flds':','.join(flds),
         'fragezeichen':fragezeichen}
      self.cursor.execute(stm,values)

   def count(self,where=None):
      """Liefert die Anzahl der Datensaetze ggf. basierend auf where"""
      tablename   = self.meta['tablename']
      
      select = 'select count(*) from %(tablename)s' % {'tablename':tablename}
      if where == None:
         self.cursor.execute(select)
      else:
         self.cursor.execute(select+' where %(where)s' % {'where':where})

      record = self.cursor.fetchone()
      return record[0]


   def sum(self,fld,where=None):
      """Liefert die Summenfunktion ggf. basierend auf where"""
      tablename   = self.meta['tablename']
      
      select = 'select sum(%(fld)s) from %(tablename)s' % {'tablename':tablename,'fld':fld}
      if where == None:
         self.cursor.execute(select)
      else:
         self.cursor.execute(select+' where %(where)s' % {'where':where})

      record = self.cursor.fetchone()
      return record[0]
   
   def min(self,fld,where=None):
      """Liefert die Summenfunktion ggf. basierend auf where"""
      tablename   = self.meta['tablename']
      
      select = 'select min(%(fld)s) from %(tablename)s' % {'tablename':tablename,'fld':fld}
      if where == None:
         self.cursor.execute(select)
      else:
         self.cursor.execute(select+' where %(where)s' % {'where':where})

      record = self.cursor.fetchone()
      return record[0]

   def max(self,fld,where=None):
      """Liefert die Summenfunktion ggf. basierend auf where"""
      tablename   = self.meta['tablename']
      
      select = 'select max(%(fld)s) from %(tablename)s' % {'tablename':tablename,'fld':fld}
      if where == None:
         self.cursor.execute(select)
      else:
         self.cursor.execute(select+' where %(where)s' % {'where':where})

      record = self.cursor.fetchone()
      return record[0]
   
   def avg(self,fld,where=None):
      """Liefert die Summenfunktion ggf. basierend auf where"""
      tablename   = self.meta['tablename']
      
      select = 'select avg(%(fld)s) from %(tablename)s' % {'tablename':tablename,'fld':fld}
      if where == None:
         self.cursor.execute(select)
      else:
         self.cursor.execute(select+' where %(where)s' % {'where':where})

      record = self.cursor.fetchone()
      return record[0]

   
   def getDomain(self,where=None) :
      'Gibt genau eine Domain oder None zurueck, wenn nicht gefunden.'
      
      getDomain = GetDomain(self)
      getDomain.get(self,where)

   def __del__(self):
      self.db.db.connection.commit()




"""
Fuer eine Tabelle wird eine Domain angelegt.

Eine Domain ist eine Klasse, welche von Domain erbt.

Beispiel einer DOmain:
----------------------

class LovDomain(Domain):
(1)lovID     = None
   lovClass  = None
   lovKey    = None
   lovValue  = None
   lovFlag1  = None
   lovFlag2  = None
   lovFlag3  = None
   lovFlag4  = None
   lovRemark = None

(2)meta = {
      'tablename':'lov',
      'primarykey':'lovID',
      
      'fields':{
         'lovID'       : {'dbfield':'lovID'},
         'lovClass'    : {'dbfield':'lovClass'},
         'lovKey'      : {'dbfield':'lovKey'},
         'lovValue'    : {'dbfield':'lovValue'},
         'lovFlag1'    : {'dbfield':'lovFlag1'},
         'lovFlag2'    : {'dbfield':'lovFlag2'},
         'lovFlag3'    : {'dbfield':'lovFlag3'},
         'lovFlag4'    : {'dbfield':'lovFlag4'},
         'lovRemark'   : {'dbfield':'lovRemark'}
         }     
      }
 (3)def getLovValueUpper(self):
      return self.lovValue.upper()

1: Die Domainfelder. Es muessen nicht alle moeglichen Datenbankfelder
   angegeben werden.

2: Die Datenstruktur, welche die Metadaten zur Datenbantabelle angeibt.
   tablename:     Name der Datenbanktabelle
   primarykey:    Datenbankname des Feldes in der Datenbanktabelle
                  welchen den Eindeutigen Schluessel enthaelt.
3: Fields
   Dies ist eine Datenstruktur die fuer jedes gewueschte Datenbankfeld
   eine Verbindung zum Domainfeld darstellt.
   Es ist moeglich, dass die Namen der Datenbankfelder und die der Domain
   unterschiedlich sind.

"""